@@ -146,7 +146,7 @@ static void a_pairing_proj(element_ptr out, element_ptr in1, element_ptr in2,
 	element_set1(z2);
     }
 
-    void proj_double()
+    void proj_double(void)
     {
 	element_ptr x = V->x;
 	element_ptr y = V->y;
