@@ -114,7 +114,7 @@ void c_param_inp_str(c_param_ptr p, FILE *stream)
     param_clear_tab(tab);
     symtab_clear(tab);
 }
-//
+
 //assumes P is in the base field, Q in some field extension
 static void cc_miller(element_t res, mpz_t q, point_t P,
 	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
@@ -275,144 +275,6 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
     element_clear(e1);
 }
 
-static void point_proj_double(point_t R, element_t Rz, point_t P, element_t z)
-{
-    common_curve_ptr cc = P->curve->data;
-    element_t e0, e1, e2, e3;
-    element_init(e0, z->field);
-    element_init(e1, z->field);
-    element_init(e2, z->field);
-    element_init(e3, z->field);
-    element_ptr x = P->x;
-    element_ptr y = P->y;
-    //e0 = 3x^2 + (cc->a) z^4
-    element_square(e0, x);
-    element_mul_si(e0, e0, 3);
-    element_square(e1, z);
-    element_square(e1, e1);
-    element_mul(e1, e1, cc->a);
-    element_add(e0, e0, e1);
-
-    //z_out = 2 y z
-    element_mul(Rz, y, z);
-    element_mul_si(Rz, Rz, 2);
-
-    //e1 = 4 x y^2
-    element_square(e2, y);
-    element_mul(e1, x, e2);
-    element_mul_si(e1, e1, 4);
-
-    //x_out = e0^2 - 2 e1
-    element_mul_si(e3, e1, 2);
-    element_square(R->x, e0);
-    element_sub(R->x, R->x, e3);
-
-    //e2 = 8y^4
-    element_square(e2, e2);
-    element_mul_si(e2, e2, 8);
-
-    //y_out = e0(e1 - x_out) - e2
-    element_sub(e1, e1, R->x);
-    element_mul(e0, e0, e1);
-    element_sub(R->y, e0, e2);
-
-    R->inf_flag = 0;
-
-    element_clear(e0);
-    element_clear(e1);
-    element_clear(e2);
-    element_clear(e3);
-}
-
-static void point_proj_mixin(point_t R, element_t Rz, point_t P, element_t Pz, point_t Q)
-{
-    element_t e2, e3, e5, e6, e9;
-    element_init(e2, Rz->field);
-    element_init(e3, Rz->field);
-    element_init(e5, Rz->field);
-    element_init(e6, Rz->field);
-    element_init(e9, Rz->field);
-    element_ptr Px = P->x;
-    element_ptr Py = P->y;
-    element_ptr Qx = Q->x;
-    element_ptr Qy = Q->y;
-
-    //e2 = Qx Pz^2, e5 holds Pz^2 until later
-    element_square(e5, Pz);
-    element_mul(e2, e5, Qx);
-
-    //e3 = Px - e2
-    element_sub(e3, Px, e2);
-
-    //e5 = Qy Pz^3 TODO: check this one
-    element_mul(e5, e5, Qy);
-    element_mul(e5, e5, Pz);
-
-    //e6 = Py - e5
-    element_sub(e6, Py, e5);
-
-    //e7 = Px + e2, use e2 to double for e7
-    element_add(e2, Px, e2);
-
-    //e8 = Py + e5, use e5 to double for e8
-    element_add(e5, Py, e5);
-
-    //Rz = Pz e3
-    element_mul(Rz, Pz, e3);
-
-    //Rx = e6^2 - e7 e3^2
-    //e3 now holds e3^3,
-    //e9 holds e7 e3^2
-    element_square(e9, e3);
-    element_mul(e3, e9, e3);
-    element_square(R->x, e6);
-    element_mul(e9, e2, e9);
-    element_sub(R->x, R->x, e9);
-
-    //e9 = e7 e3^2 - 2 Rx
-    element_sub(e9, e9, R->x);
-    element_sub(e9, e9, R->x);
-
-    //Ry = (e9 e6 - e8 e3^3)/2
-    element_mul(e9, e9, e6);
-    element_mul(e5, e5, e3);
-    element_sub(e9, e9, e5);
-    //TODO: rewrite so it doesn't access internal structure
-    {
-	mpz_ptr z0 = e9->data;
-	if (mpz_odd_p(z0)) {
-	    mpz_add(z0, z0, e9->field->order);
-	}
-	mpz_tdiv_q_2exp(R->y->data, z0, 1);
-    }
-
-    R->inf_flag = 0;
-
-    element_clear(e2);
-    element_clear(e3);
-    element_clear(e5);
-    element_clear(e6);
-    element_clear(e9);
-}
-
-//convert P from weighted projective (Jacobian) to affine
-//i.e. (X, Y, Z) --> (X/Z^2, Y/Z^3)
-//also sets z to 1
-static void point_to_affine(point_ptr P, element_t z)
-{
-    element_t e0;
-
-    element_init(e0, z->field);
-    element_invert(z, z);
-    element_square(e0, z);
-    element_mul(P->x, P->x, e0);
-    element_mul(e0, e0, z);
-    element_mul(P->y, P->y, e0);
-    element_set1(z);
-
-    element_clear(e0);
-}
-
 static void cc_tatepower(element_ptr out, element_ptr in, pairing_t pairing)
 {
     mnt_pairing_data_ptr p = pairing->data;
@@ -539,15 +401,108 @@ static void pairing_init_c_param_odd_k(pairing_t pairing, c_param_t param)
 static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
 {
-    //collate divisions
     int m;
     element_t v;
     point_t Z;
     element_t a, b, c;
     common_curve_ptr cc = P->curve->data;
     element_t t0, t1;
+    element_ptr t2 = a, t3 = b, t4 = c;
     element_t e0, e1;
-    element_t z;
+    element_t z, z2;
+
+    void proj_double(void)
+    {
+	element_ptr x = Z->x;
+	element_ptr y = Z->y;
+	//t0 = 3x^2 + (cc->a) z^4
+	element_square(t0, x);
+	element_mul_si(t0, t0, 3);
+	element_square(t1, z2);
+	element_mul(t1, t1, cc->a);
+	element_add(t0, t0, t1);
+
+	//z_out = 2 y z
+	element_mul(z, y, z);
+	element_mul_si(z, z, 2);
+	element_square(z2, z);
+
+	//t1 = 4 x y^2
+	element_square(t2, y);
+	element_mul(t1, x, t2);
+	element_mul_si(t1, t1, 4);
+
+	//x_out = t0^2 - 2 t1
+	element_mul_si(t3, t1, 2);
+	element_square(x, t0);
+	element_sub(x, x, t3);
+
+	//t2 = 8y^4
+	element_square(t2, t2);
+	element_mul_si(t2, t2, 8);
+
+	//y_out = t0(t1 - x_out) - t2
+	element_sub(t1, t1, x);
+	element_mul(t0, t0, t1);
+	element_sub(y, t0, t2);
+    }
+
+    void proj_mixin(void)
+    {
+	element_ptr Px = Z->x;
+	element_ptr Py = Z->y;
+	element_ptr Qx = P->x;
+	element_ptr Qy = P->y;
+
+	//t2 = Qx z^2
+	element_mul(t2, z2, Qx);
+
+	//t3 = Px - t2
+	element_sub(t3, Px, t2);
+
+	//t0 = Qy z^3
+	element_mul(t0, z2, Qy);
+	element_mul(t0, t0, z);
+
+	//t1 = Py - t0
+	element_sub(t1, Py, t0);
+
+	//e7 = Px + t2, use t2 to double for e7
+	element_add(t2, Px, t2);
+
+	//e8 = Py + t0, use t0 to double for e8
+	element_add(t0, Py, t0);
+
+	//z = z t3
+	element_mul(z, z, t3);
+	element_square(z2, z);
+
+	//Px = t1^2 - e7 t3^2
+	//t3 now holds t3^3,
+	//t4 holds e7 t3^2
+	element_square(t4, t3);
+	element_mul(t3, t4, t3);
+	element_square(Px, t1);
+	element_mul(t4, t2, t4);
+	element_sub(Px, Px, t4);
+
+	//t4 = e7 t3^2 - 2 Px
+	element_sub(t4, t4, Px);
+	element_sub(t4, t4, Px);
+
+	//Py = (t4 t1 - e8 t3^3)/2
+	element_mul(t4, t4, t1);
+	element_mul(t0, t0, t3);
+	element_sub(t4, t4, t0);
+	//TODO: rewrite so it doesn't access internal structure
+	{
+	    mpz_ptr z0 = t4->data;
+	    if (mpz_odd_p(z0)) {
+		mpz_add(z0, z0, t4->field->order);
+	    }
+	    mpz_tdiv_q_2exp(Py->data, z0, 1);
+	}
+    }
 
     void do_vertical(void)
     {
@@ -564,33 +519,23 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_ptr Zx = Z->x;
 	element_ptr Zy = Z->y;
 
-	if (point_is_inf(Z)) {
-	    return;
-	}
-
-	if (element_is0(Zy)) {
-	    //order 2
-	    do_vertical();
-	    return;
-	}
 	//a = -(3x^2 + cca z^4)
 	//b = 2 y z^3
 	//c = -(2 y^2 + x a)
 	//a = z^2 a
-	element_square(t0, z);
-	element_square(a, t0);
+	element_square(a, z2);
 	element_mul(a, a, cc->a);
 	element_square(b, Zx);
 	element_mul_si(b, b, 3);
 	element_add(a, a, b);
 	element_neg(a, a);
 
-	element_mul(b, z, t0);
+	element_mul(b, z, z2);
 	element_mul(b, b, Zy);
 	element_mul_si(b, b, 2);
 
 	element_mul(c, Zx, a);
-	element_mul(a, a, t0); //frees e0, don't need z^2 anymore
+	element_mul(a, a, z2);
 	element_square(t0, Zy);
 	element_mul_si(t0, t0, 2);
 	element_add(c, c, t0);
@@ -614,24 +559,12 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
 	element_ptr Bx = P->x;
 	element_ptr By = P->y;
 
-	//we assume B is never O
-	if (point_is_inf(Z)) {
-	    do_vertical();
-	    return;
-	}
-
-	if (!element_cmp(Ax, Bx)) {
-	    if (!element_cmp(Ay, By)) {
-		do_tangent();
-		return;
-	    }
-	    do_vertical();
-	    return;
-	}
+	//a = -(By z^3 - Ay)
+	//b = Bx z^3 - Ax z
+	//c = Ax z By - Ay Bx;
 
 	element_mul(t0, Ax, z);
-	element_mul(t1, z, z);
-	element_mul(t1, t1, z);
+	element_mul(t1, z2, z);
 
 	element_mul(a, By, t1);
 	element_sub(a, Ay, a);
@@ -656,12 +589,14 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
     element_init(a, P->curve->field);
     element_init(b, P->curve->field);
     element_init(c, P->curve->field);
-    element_init(t0, P->curve->field);
-    element_init(t1, P->curve->field);
+    element_init(t0, a->field);
+    element_init(t1, a->field);
     element_init(e0, res->field);
     element_init(e1, res->field);
     element_init(z, a->field);
+    element_init(z2, a->field);
     element_set1(z);
+    element_set1(z2);
 
     element_init(v, res->field);
     point_init(Z, P->curve);
@@ -671,15 +606,16 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
     element_set1(v);
     m = mpz_sizeinbase(q, 2) - 2;
 
-    while(m >= 0) {
-	element_square(v, v);
+    for(;;) {
 	do_tangent();
-	point_proj_double(Z, z, Z, z);
+	if (!m) break;
+	proj_double();
 	if (mpz_tstbit(q, m)) {
 	    do_line();
-	    point_proj_mixin(Z, z, Z, z, P);
+	    proj_mixin();
 	}
 	m--;
+	element_square(v, v);
     }
 
     element_set(res, v);
@@ -694,26 +630,23 @@ static void cc_miller_no_denom_proj(element_t res, mpz_t q, point_t P,
     element_clear(e0);
     element_clear(e1);
     element_clear(z);
+    element_clear(z2);
 }
 
 static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
 {
-    //collate divisions
     int m;
     element_t v;
     point_t Z;
     element_t a, b, c;
-    common_curve_ptr cc = P->curve->data;
+    const common_curve_ptr cc = P->curve->data;
+    const element_ptr cca = cc->a;
     element_t t0;
     element_t e0, e1;
-    element_t z;
 
     void do_vertical(void)
     {
-	if (point_is_inf(Z)) {
-	    return;
-	}
 	mapbase(e0, Z->x);
 	element_sub(e0, Qx, e0);
 	element_mul(v, v, e0);
@@ -721,32 +654,15 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 
     void do_tangent(void)
     {
-	//a = -slope_tangent(Z.x, Z.y);
-	//b = 1;
-	//c = -(Z.y + a * Z.x);
-	//but we multiply by 2*Z.y to avoid division
-
-	//a = -Zx * (Zx + Zx + Zx + twicea_2) - a_4;
-	//Common curves: a2 = 0 (and cc->a is a_4), so
 	//a = -(3 Zx^2 + cc->a)
 	//b = 2 * Zy
 	//c = -(2 Zy^2 + a Zx);
-	element_ptr Zx = Z->x;
-	element_ptr Zy = Z->y;
-
-	if (point_is_inf(Z)) {
-	    return;
-	}
-
-	if (element_is0(Zy)) {
-	    //order 2
-	    do_vertical();
-	    return;
-	}
+	const element_ptr Zx = Z->x;
+	const element_ptr Zy = Z->y;
 
 	element_square(a, Zx);
 	element_mul_si(a, a, 3);
-	element_add(a, a, cc->a);
+	element_add(a, a, cca);
 	element_neg(a, a);
 
 	element_add(b, Zy, Zy);
@@ -774,25 +690,10 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
 	//c = -(A.y + a * A.x);
 	//but we'll multiply by B.x - A.x to avoid division
 
-	element_ptr Ax = Z->x;
-	element_ptr Ay = Z->y;
-	element_ptr Bx = P->x;
-	element_ptr By = P->y;
-
-	//we assume B is never O
-	if (point_is_inf(Z)) {
-	    do_vertical();
-	    return;
-	}
-
-	if (!element_cmp(Ax, Bx)) {
-	    if (!element_cmp(Ay, By)) {
-		do_tangent();
-		return;
-	    }
-	    do_vertical();
-	    return;
-	}
+	const element_ptr Ax = Z->x;
+	const element_ptr Ay = Z->y;
+	const element_ptr Bx = P->x;
+	const element_ptr By = P->y;
 
 	element_sub(b, Bx, Ax);
 	element_sub(a, Ay, By);
@@ -817,8 +718,6 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_init(t0, P->curve->field);
     element_init(e0, res->field);
     element_init(e1, res->field);
-    element_init(z, a->field);
-    element_set1(z);
 
     element_init(v, res->field);
     point_init(Z, P->curve);
@@ -828,15 +727,18 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_set1(v);
     m = mpz_sizeinbase(q, 2) - 2;
 
-    while(m >= 0) {
-	element_square(v, v);
+    for(;;) {
 	do_tangent();
+
+	if (!m) break;
+
 	point_double(Z, Z);
 	if (mpz_tstbit(q, m)) {
 	    do_line();
 	    point_add(Z, Z, P);
 	}
 	m--;
+	element_square(v, v);
     }
 
     element_set(res, v);
@@ -849,7 +751,6 @@ static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
     element_clear(t0);
     element_clear(e0);
     element_clear(e1);
-    element_clear(z);
 }
 
 static void cc_tatepower_even_k(element_ptr out, element_ptr in, pairing_t pairing)
